/*
  DXC Framework (http://dx-competition.org/)
  Copyright 2008 David Garcia

  The DXC Framework is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the
  Free Software Foundation, either version 3 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program. If not, see <http://www.gnu.org/licenses/>
*/
#include <string.h>
#include <sstream>
#include <iostream>
#include <map>
#include <ctime>
#include <assert.h>
#include <errno.h>
#include <stdlib.h>

#ifndef _WIN32
# include <sys/poll.h>
#else
#include <winsock2.h>
#include <ws2tcpip.h>
#endif

#include <DxcCallback.hpp>
#include <Connection/DxcConnectorImp.hpp>
#include <Parser/MessageDriver.hpp>

using namespace Dxc;
using namespace std;

#ifdef _WIN32

#pragma warning (disable:4996)

typedef int socklen_t;

int initSock()
{
    WSADATA wsd;
    WORD version = MAKEWORD(2, 2);

    int err = WSAStartup(version, &wsd);
    if (err != 0)
    {
        string e("WSAStartup failed with error: ");
        e += err;
        throw DxcException(e);
    }

    return err;
}

bool recvBufferTooSmall()
{
    if (WSAGetLastError() == WSAEMSGSIZE)
        return true;
    else
        return false;
}

bool DxcConnectorImp::socketWriteReady(int sockfd)
{
    fd_set fds;

    fds.fd_count = 1;
    fds.fd_array[0] = sockfd;

    ::timeval tv;
    tv.tv_sec = 0;
    tv.tv_usec = 300000;

    int stat = select(1, NULL, &fds, NULL, &tv);

    if (stat == 1)
        return true;

    return false;
}

bool DxcConnectorImp::socketReadReady(int sockfd)
{
    fd_set fds;

    FD_ZERO(&fds);
    FD_SET(sockfd, &fds);

    ::timeval tv;
    tv.tv_sec = 0;
    tv.tv_usec = 5000;

    int stat = select(1, &fds, NULL, NULL, &tv);

    if (stat == 1)
    {
        return true;
    }

    return false;
}

#else

#include <arpa/inet.h>
#include <sys/socket.h>
#include <unistd.h>
#include <netdb.h>

int initSock()
{
    return 0;
}

bool recvBufferTooSmall()
{
    /*
        if (WSAGetLastError() == WSAEMSGSIZE)
            return true;
        else
            return false;
    */
    return false;
}

bool DxcConnectorImp::socketWriteReady(int sockfd)
{
    pollfd pfd;
    pfd.fd = sockfd;
    pfd.events = POLLOUT;
    int stat = poll(&pfd, 1, 50);

    if (stat > 0)
    {
        short r = pfd.revents;
        if (r & POLLHUP)
        {
//            cout << "poll_socket(): poll() detected hangup (POLLHUP)" << endl;
            return false;
        }
        else if (r & POLLERR)
        {
//            cout << "poll_socket(): poll() POLLERR" << endl;
            return false;
        }
        else if (r & POLLNVAL)
        {
//            cout <<"poll_socket(): poll() POLLNVAL" << endl;
            return false;
        }
        else
        {
            return true;
        }
    }
}

bool DxcConnectorImp::socketReadReady(int sockfd)
{
    pollfd pfd;
    pfd.fd = sockfd;
    pfd.events = POLLIN | POLLPRI;
    pfd.revents = 0;

    int stat = poll(&pfd, 1, 50);

    if (stat > 0)
    {
        short r = pfd.revents;
        if (r & POLLHUP)
        {
//            cout << "poll_socket(): poll() detected hangup (POLLHUP)" << endl;
            return false;
        }
        else if (r & POLLERR)
        {
//            cout << "poll_socket(): poll() POLLERR" << endl;
            return false;
        }
        else if (r & POLLNVAL)
        {
//            cout <<"poll_socket(): poll() POLLNVAL" << endl;
            return false;
        }
        else
        {
            return true;
        }
    }

    return true;
}
#endif

DxcConnectorImp::DxcConnectorImp(DxcCallback * cb, const string& logFilePrefix)
    : callback_(cb), logFile_(NULL), shutdown_(false)
{
    initSock();
    pthread_mutex_init(&parsingMutex_, NULL);

    if (logFilePrefix != "")
    {
        time_t rawtime;
        tm* t;
        time(&rawtime);
        t = localtime(&rawtime);

        ostringstream nameStr;
        nameStr << logFilePrefix;
        nameStr << "-" << (1900 + t->tm_year)
                << "-" << t->tm_mon << "-" << t->tm_mday << "-"
                << t->tm_hour << t->tm_min << ".log";

        logFile_ = new ofstream(nameStr.str().c_str());
        *logFile_ << "foo;" << endl;
        logFile_->close();
    }
}

void DxcConnectorImp::acceptConnection(int port, int *sockp)
{
    acceptConnection("", port, sockp);
}


void DxcConnectorImp::acceptConnection(string source, int port, int *sockp)
{
    if ((listenfd_ = socket(PF_INET, SOCK_STREAM, 0)) < 0)
    {
        ostringstream e;
        e << "Couldn't connect on port " << port;
        throw DxcException(e.str());
    }
#ifdef _WIN32
    const char on = 1;
#else
    int on = 1;
#endif
    if (setsockopt(listenfd_, SOL_SOCKET, SO_REUSEADDR, &on,  sizeof(on)))
    {
        ostringstream e;
        e << "Couldn't set bind socket options. Socket: " << listenfd_;
        throw DxcException(e.str());
    }

    struct sockaddr_in serv;

    memset(&serv, 0, sizeof(struct sockaddr_in));

    serv.sin_family = AF_INET;
    serv.sin_addr.s_addr = htonl(INADDR_ANY);
    serv.sin_port = htons(port);

    if (bind(listenfd_, (struct sockaddr *)&serv, sizeof(serv)) < 0)
    {
        ostringstream e;
        e << "Couldn't bind on port " << port;
        throw DxcException(e.str());
    }
    if (listen(listenfd_, SOMAXCONN) < 0)
    {
        ostringstream e;
        e << "Couldn't listen on port " << port;
        throw DxcException(e.str());
    }
    /*
        cout << "socket: " << listenfd_ << "Listening on port: " << port << endl;
    */
    ReceiveParam * rp = new ReceiveParam;
    rp->sockfdp = sockp;
    *(rp->sockfdp) = listenfd_;
    rp->cb = callback_;
    rp->source = new string(source);
    rp->shutdown = &shutdown_;
    rp->parsingMutex = &parsingMutex_;

    pthread_t *recvThread = new pthread_t;

    int threadRet = pthread_create(recvThread,
                                   NULL,
                                   DxcConnectorImp::acceptAndReceive,
                                   rp);

    if (threadRet)
    {
        delete recvThread, rp;
        throw DxcException("Unable to create recvThread.");
    }

    recvThreads_.push_back(recvThread);
}

int DxcConnectorImp::connect(string host, int port, int *sockp)
{
    return connect("", host, port, sockp);
}

int DxcConnectorImp::connect(string source, string host, int port, int *sockp)
{
    int tmpsock;

    if ((tmpsock = socket(PF_INET, SOCK_STREAM, 0)) < 0)
    {
        string e("Couldn't create socket.");
        throw DxcException(e);
    }

    struct addrinfo hints, *res;
    int err;
    memset(&hints, 0, sizeof(hints));
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_family = AF_INET;

    ostringstream oss;
    oss << port;
    if ((err = getaddrinfo(host.c_str(), oss.str().c_str(), &hints, &res)) != 0)
    {
        string e("Couldn't find hostname.");
        cout << "Couldn't find hostname. " << oss.str().c_str() << endl;
        throw DxcException(e);
    }

    if (::connect(tmpsock, res->ai_addr, res->ai_addrlen) < 0)
    {
        ostringstream e;
        e << "Couldn't connect on port " << port;
        throw DxcException(e.str());
    }

    *sockp = tmpsock;

    ReceiveParam *rp = new ReceiveParam;
    rp->sockfdp = sockp;
    rp->source = new string(source);
    rp->cb = callback_;
    rp->shutdown = &shutdown_;
    rp->parsingMutex = &parsingMutex_;

    pthread_t *recvThread = new pthread_t;

    int threadRet = pthread_create(recvThread,
                                   NULL,
                                   DxcConnectorImp::receive,
                                   rp);

    if (threadRet)
    {
        string e("Couldn't create recvThread.");
        throw DxcException(e);
    }

    recvThreads_.push_back(recvThread);

    return *sockp;
}

bool DxcConnectorImp::transmit(int sockfd, const DxcData *msg)
{
    if (sockfd < 0)
    {
        return false;
    }
    else if (!socketWriteReady(sockfd))
    {
        return false;
    }

    ostringstream s;
    s << msg;
    string msgStr(s.str());

    if (send(sockfd, msgStr.c_str(), msgStr.length(), 0x0) < 0)
    {
        return false;
    }

    if (logFile_ && logFile_->is_open())
    {
        *logFile_ << msgStr << endl;
        logFile_->flush();
    }
    return true;
}

void *DxcConnectorImp::receive(void *receiveParam)
{
    ReceiveParam *rp = static_cast<ReceiveParam *>(receiveParam);

    if (rp->cb == NULL || *(rp->sockfdp) < 0)
    {
        delete rp;
        return NULL;
    }

    DxcCallback *callback = rp->cb;
    string source(*(rp->source));
    string cbtype(typeid(*callback).name());
    int sockfd = *(rp->sockfdp);
    pthread_mutex_t* parsingMutex = rp->parsingMutex;

    size_t len = 4096;
    struct timeval tv;

    tv.tv_sec = 0;
    tv.tv_usec = 500;

    int readRet = 1;
    MessageDriver md;
    string strBuf = "";

    char *buf = (char *)malloc(len * sizeof(char));
    if (NULL == buf)
    {
        cout << "error allocating memory." << endl;
        return NULL;
    }

    if (setsockopt(sockfd, SOL_SOCKET, SO_RCVTIMEO, (char *)&tv,  sizeof tv))
    {
        cout << "receive() couldn't set socket options. Socket: " << sockfd << endl;
        return NULL;
    }

    while ( !(*(rp->shutdown)) )
    {
        if (socketReadReady(sockfd))
        {
            readRet = recv(sockfd, buf, len, 0);
            if (recvBufferTooSmall())
            {
                char *p = (char *)realloc(buf, len * 2 * sizeof(char));
                if (NULL == p)
                {
                    cout << "error allocating memory." << endl;
                    return NULL;
                }
                len *= 2;
                buf = p;
            }
            else if (readRet > 0)
            {
                string::size_type n;
                strBuf.append(buf, readRet);

                n = strBuf.find_last_of(';');

                if (n < strBuf.length())
                {
                    string tmpStr(strBuf, 0, n + 1);
                    strBuf.erase(0, n + 1);

                    try
                    {
                        pthread_mutex_lock(parsingMutex);
                        md.parseString(tmpStr);
                        pthread_mutex_unlock(parsingMutex);
                    }
                    catch (DxcException ex)
                    {
                        ostringstream err;
                        err << ex.what() << ". Attempting to parse: " << tmpStr;
                        cout << err.str();

                        ErrorData ed(DxcData::currentTime(), err.str());
                        callback->onDataReceived(&ed);
                        pthread_mutex_unlock(parsingMutex);
                    }

                    if (md.hasMessages())
                    {
                        DxcData* m = NULL;
                        while (m = md.getNextMessage())
                        {
                            if (callback)
                            {
                                if (source.length() > 0)
                                    m->setSource(source);
                                callback->onDataReceived(m);
                            }
                            delete m;
                        }
                    }
                }
            }
        }
    }

    closeSocket(sockfd);

    free(buf);

    delete rp;

    return NULL;
}

void *DxcConnectorImp::acceptAndReceive(void *receiveParam)
{
    ReceiveParam * rp = static_cast<ReceiveParam *>(receiveParam);

    if (rp->cb == NULL || *(rp->sockfdp) < 0)
    {
        delete rp;
        return NULL;
    }

    int sockfd = *(rp->sockfdp);
    *(rp->sockfdp) = -1; // so that external thread doesn't send/rcv until
    // the connection is established
    struct sockaddr_in child;

    memset(&child, 0, sizeof(child));
    socklen_t childlen = (socklen_t)sizeof(child);
    while (!socketReadReady(sockfd) && !(*(rp->shutdown)))
    {
    }

    if (!(*(rp->shutdown)) && socketReadReady(sockfd) &&
            (sockfd = accept(sockfd, (struct sockaddr *)&child, &childlen)) >= 0)
    {
        (*rp->sockfdp) = sockfd;

        receive(rp);
    }
    else
    {
        cout << "Couldn't accept connection. Socket: " << sockfd;
        return NULL;
    }


    return NULL;
}

DxcConnectorImp::~DxcConnectorImp()
{
    shutdown_ = true;

    // join threads and delete pointers in recvThreads_
    while (!recvThreads_.empty())
    {
        pthread_t *tp = recvThreads_.back();
        pthread_join(*tp, NULL);
        delete tp;

        recvThreads_.pop_back();
    }

    if (logFile_ && logFile_->is_open())
    {
        logFile_->close();
    }

    closeSocket(listenfd_);
}

void DxcConnectorImp::closeSocket(int socket)
{
    int ret = shutdown(socket, 1);
#ifdef _WIN32
    closesocket(socket);

#pragma warning (default:4996)
#else
    close(socket);
#endif
}
