/*
  DXC Framework (http://dx-competition.org/)
  Copyright 2008 David Garcia

  The DXC Framework is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the
  Free Software Foundation, either version 3 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program. If not, see <http://www.gnu.org/licenses/>
*/
#ifndef DIAGNOSIS_DATA_HPP
#define DIAGNOSIS_DATA_HPP

#include <set>
#include <map>
#include <vector>
#include <DxcDataTypes.hpp>
#include <DxcDefs.hpp>

namespace Dxc
{
/**
 * @brief Defines the data structure for a Diagnosis.
 *
 * A diagnosis is a set of candidates, weighed by their relative probability.
 * The weights of these candidates are normalized upon construction of the
 * datatype, resulting in weights that yield 1 when added together.
 * An additional detection signal and isolation signal indicate whether the
 * reasoner has detected something wrong with the system (detectionSignal =
 * true) and wether it has isolated the fault to specific component failure
 * modes (isolationSignal = true). If the isolationSignal is false, the
 * Candidate set, regardless of its contents, may be effectively treated
 * as empty.
 */
class DXCAPI_EXPORT DiagnosisData : public DxcData
{
public:

    /**
     * @brief Constructor to initialize the DiagnosisData with timestamp.
     *
     * @param timestamp - message timestamp.
     * @param detectionSignal - set to true if the system is believed to be in a
     *             faulty state, even if there is not yet an isolation (CandidateSet).
     * @param isolationSignal - set to true if faults have been isolated, ie. candidates exist.
     * @param isolation - a set of Candidates (hypotheses). Their weights will be
     *            normalized, so that they all add up to one.
     * @param notes - A string for Diagnosis Algorithms to supply additional information for their own
     *            purposes.
     */
    DiagnosisData(Timestamp timestamp,
                  bool detectionSignal = false,
                  bool isolationSignal = false,
                  const CandidateSet& isolation = CandidateSet(),
                  const std::string& notes = "");

    /**
     * @brief Constructor to initialize the DiagnosisData with current time as timestamp.
     *
     * @param detectionSignal - set to true if the system is believed to be in a
     *             faulty state, even if there is not yet an isolation (CandidateSet).
     * @param isolationSignal - set to true if faults have been isolated, ie. candidates exist.
     * @param isolation - a set of Candidates (hypotheses). Their weights will be
     *            normalized, so that they all add up to one.
     * @param notes - A string for Diagnosis Algorithms to supply additional information for their own
     *            purposes.
     */
    DiagnosisData(bool detectionSignal = false,
                  bool isolationSignal = false,
                  const CandidateSet& isolation = CandidateSet(),
                  const std::string& notes = "");

    // this constructor is meant to be called from the parser
    DiagnosisData(Timestamp timestamp,
                  bool detectionSignal,
                  bool isolationSignal,
                  const std::vector<FaultSet*>& candidates,
                  const std::vector<double>& weights,
                  const std::string& notes);

    /**
     * @brief True if, according to the diagnosis, the system is believed to be in a
     *        faulty state.
     */
    bool getDetectionSignal() const;

    /**
     * @brief True if faults have been isolated, ie. candidates exist.
     */
    bool getIsolationSignal() const;

    const CandidateSet* getCandidates() const;
    std::string getNotes() const;

    virtual DiagnosisData* clone() const;

    virtual std::ostream &put(std::ostream &) const;

private:
    bool detectionSignal_;
    bool isolationSignal_;
    CandidateSet candidateSet_;
    std::string notes_;
};
}

#endif

