/*
  DXC Framework (http://dx-competition.org/)
  Copyright 2008 David Garcia

  The DXC Framework is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the
  Free Software Foundation, either version 3 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program. If not, see <http://www.gnu.org/licenses/>
*/
#ifndef _WIN32

#include <errno.h>
#include <string.h>
#include <signal.h>

#include <cstdlib>
#include <string>
#include <vector>
#include <iostream>

#include <sys/wait.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#include <DataTypes/DxcException.hpp>
#include <DataTypes/DxcData.hpp>
#include <LinuxExeLoader.hpp>

using namespace std;
using namespace Dxc;

string nextToken(const string& str, size_t * pos);

LinuxExeLoader::LinuxExeLoader(const string &exeName,
                               const string &path,
                               const string &args)
    : exe_(exeName), path_(path), args_(args)
{
}

void LinuxExeLoader::start()
{
    char **args;
    string nameStr = path_ + DIR_DELIM + exe_;
    size_t argsLen = args_.length();
    size_t pos = 0, apos = 0;

    string tk = nextToken(args_, &pos);
    vector<string> tkv;

    pid_ = -1;

    tkv.push_back(nameStr);
    while (tk != "")
    {
        tkv.push_back(tk);
        tk = nextToken(args_, &pos);
    }

    args = new char *[tkv.size() + 1];

    size_t i;
    for (i = 0; i < tkv.size(); i++)
    {
        args[i] = new char[tkv.at(i).length() + 1];
        tkv.at(i).copy(args[i], tkv.at(i).length());
        args[i][tkv.at(i).length()] = '\0';
    }

    args[i] = '\0';

    times(&startTime);

    char *dxcHome = getenv("DXC_HOME");
    if (NULL == dxcHome)
    {
        throw DxcException(getError().c_str());
    }
    string dxcMemProfLib = string(dxcHome) + "/Lib/libdxcMemProf.so";
    setenv("LD_PRELOAD", dxcMemProfLib.c_str(), 1 /* overwrite */);

    key_ = ftok(dxcMemProfLib.c_str(), 0);
    if (key_ == -1)
    {
        throw DxcException(getError().c_str());
    }


    //    signal(SIGCLD, SIG_IGN); // ignore SIGCLD so we don't have to wait()
    if ((pid_ = fork()) == 0)
    {
        // Child
        shmid_ = shmget(key_, sizeof(long), 0644 | IPC_CREAT);
        if (shmid_ == -1)
        {
            throw DxcException(getError().c_str());
        }
        long *memuse = (long *)shmat(shmid_, NULL, 0);
        if (memuse == (void *)-1)
        {
            throw DxcException(getError().c_str());
        }
        *memuse = 0;
        shmdt(memuse);

        execv(nameStr.c_str(), args);
        throw DxcException(getError().c_str());
    }
    else if (pid_ == -1)
    {
        throw DxcException(getError().c_str());
    }
    // Parent

    for (i = 0; i < tkv.size(); i++)
    {
        delete [] args[i];
    }

    delete [] args;
}

unsigned long LinuxExeLoader::getCpuTime()
{
    long tps = sysconf(_SC_CLK_TCK);

    times(&endTime);

    unsigned long result = ((endTime.tms_cutime - startTime.tms_cutime) +
                            (endTime.tms_cstime - startTime.tms_cstime)) * 1000 / tps;

    return result;
}

bool LinuxExeLoader::isRunning()
{
    int status;
    if (kill(pid_, 0) == 0)
    {
        return true;
    }

    return false;
}

bool LinuxExeLoader::terminate()
{
    int status;

//    if (fork() == 0)
//    {
    kill(pid_, SIGKILL);

//    }

    waitpid(pid_, &status, WNOHANG);

    return true;
}

bool LinuxExeLoader::timedWait(long time)
{
    int status;

    Timestamp t = DxcData::currentTime();
    while (waitpid(pid_, &status, WNOHANG) == 0/*isRunning()*/)
    {
        Timestamp t1 = DxcData::currentTime();
        long dt = static_cast<long>(t1 - t);
        t = t1;

        long st = 1000 - dt;
        msleep((st > 0) ? st : 0);
        t1 = DxcData::currentTime();
        dt = static_cast<long>(t1 - t);
        time -= dt;
        t = DxcData::currentTime();
        if (time < 0)
        {
            return false;
        }
    }
    return true;
}

string LinuxExeLoader::getError()
{
    return strerror(errno);
}

LinuxExeLoader::~LinuxExeLoader()
{
    int status;

    if (isRunning())
    {
//        wait(&status);

        terminate();

    }
}

string nextToken(const string& str, size_t* pos)
{
    if (*pos == string::npos || *pos >= str.length())
        return "";

    size_t n = str.find_first_of("\"' ", *pos);
    if (n == string::npos)
    {
        size_t tmp = *pos;
        *pos = n;
        return str.substr(tmp);
    }

    size_t m;
    if (str[n] == '\'')
    {
        m = str.find_first_of('\'', n + 1);
    }
    else if (str[n] == '"')
    {
        m = str.find_first_of('"', n + 1);
    }
    else
    {
        if (n != 0)
        {
            if (str[n-1] == '\\')
            {
                size_t tmp = *pos;
                *pos = n + 1;
                return str.substr(tmp, n - tmp - 1) + " " + nextToken(str, pos);
            }
        }
        size_t tmp = *pos;
        *pos = n + 1;
        return str.substr(tmp, n - tmp);
    }

    if (m == string::npos)
    {
        size_t tmp = *pos;
        *pos = m;
        string tk = str.substr(tmp);
        return tk;
    }
    else
    {
        size_t tmp = *pos;
        *pos = m + 1;
        string tk = str.substr(tmp, m - tmp + 1) + nextToken(str, pos);
        return tk;
    }

    return "";
}

unsigned long LinuxExeLoader::getPeakMemoryUsage()
{
    shmid_ = shmget(key_, sizeof(long), 0644 | IPC_CREAT);
    if (shmid_ == -1)
    {
        throw DxcException(getError().c_str());
    }
    long *memuse = (long *)shmat(shmid_, NULL, 0);
    if (memuse == (void *)-1)
    {
        throw DxcException(getError().c_str());
    }
    unsigned long result = *memuse;
    shmdt(memuse);

    shmctl(shmid_, IPC_RMID, NULL);

    return result;
}

#endif
